package javax.naming.ldap;

import java.util.Locale;
import java.io.StreamCorruptedException;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import javax.naming.directory.BasicAttributes;
import java.util.Collections;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.InvalidNameException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import java.util.ArrayList;
import java.io.Serializable;

public class Rdn implements Serializable, Comparable<Object>
{
    private transient ArrayList<Rdn.RdnEntry> entries;
    private static final int DEFAULT_SIZE = 1;
    private static final long serialVersionUID = -5994465067210009656L;
    private static final String escapees = ",=+<>#;\"\\";
    
    public Rdn(Attributes attributes) throws InvalidNameException {
        if (attributes.size() == 0) {
            throw new InvalidNameException("Attributes cannot be empty");
        }
        this.entries = new ArrayList(attributes.size());
        NamingEnumeration<Object> all = attributes.getAll();
        try {
            for (int index = 0; all.hasMore(); ++index) {
                Rdn.RdnEntry element = new Rdn.RdnEntry();
                Attribute attribute = (Attribute)all.next();
                element.type = attribute.getID();
                element.value = attribute.get();
                this.entries.add(index, element);
            }
        }
        catch (NamingException cause) {
            InvalidNameException o = new InvalidNameException(cause.getMessage());
            o.initCause(cause);
            throw o;
        }
        this.sort();
        return;
    }
    
    public Rdn(String s) throws InvalidNameException {
        this.entries = new ArrayList(1);
        new Rfc2253Parser(s).parseRdn(this);
    }
    
    public Rdn(Rdn rdn) {
        (this.entries = new ArrayList(rdn.entries.size())).addAll(rdn.entries);
    }
    
    public Rdn(String str, Object obj) throws InvalidNameException {
        if (obj == null) {
            throw new NullPointerException("Cannot set value to null");
        }
        if (str.equals("") || this.isEmptyValue(obj)) {
            throw new InvalidNameException("type or value cannot be empty, type:" + str + " value:" + obj);
        }
        this.entries = new ArrayList(1);
        this.put(str, obj);
        return;
    }
    
    private boolean isEmptyValue(Object o) {
        return (o instanceof String && o.equals("")) || (o instanceof byte[] && ((byte[])(byte[])o).length == 0);
    }
    
    Rdn() {
        this.entries = new ArrayList(1);
    }
    
    Rdn put(String s, Object o) {
        Rdn.RdnEntry e = new Rdn.RdnEntry();
        e.type = s;
        if ((o instanceof byte[])) {
            e.value = ((byte[])(byte[])o).clone();
        }
        else {
            e.value = o;
        }
        this.entries.add(e);
        return this;
    }
    
    void sort() {
        if (this.entries.size() > 1) {
            Collections.sort(this.entries);
        }
    }
    
    public Object getValue() {
        return ((Rdn.RdnEntry)this.entries.get(0)).getValue();
    }
    
    public String getType() {
        return ((Rdn.RdnEntry)this.entries.get(0)).getType();
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        int size = this.entries.size();
        if (size > 0) {
            sb.append(this.entries.get(0));
        }
        for (int i = 1; i < size; ++i) {
            sb.append('+');
            sb.append(this.entries.get(i));
        }
        return sb.toString();
    }
    
    @Override
    public int compareTo(Object o) {
        if (!(o instanceof Rdn)) {
            throw new ClassCastException("The obj is not a Rdn");
        }
        if (o == this) {
            return 0;
        }
        Rdn rdn = (Rdn)o;
        int min = Math.min(this.entries.size(), rdn.entries.size());
        int i = 0;
        while (i < min) {
            int compareTo = ((Rdn.RdnEntry)this.entries.get(i)).compareTo((Rdn.RdnEntry)rdn.entries.get(i));
            if (compareTo != 0) {
                return compareTo;
            }
            ++i;
        }
        return this.entries.size() - rdn.entries.size();
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof Rdn)) {
            return false;
        }
        Rdn rdn = (Rdn)o;
        if (this.entries.size() != rdn.size()) {
            return false;
        }
        int i = 0;
        while (i < this.entries.size()) {
            if (!((Rdn.RdnEntry)this.entries.get(i)).equals(rdn.entries.get(i))) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    @Override
    public int hashCode() {
        int n = 0;
        for (int i = 0; i < this.entries.size(); ++i) {
            n += ((Rdn.RdnEntry)this.entries.get(i)).hashCode();
        }
        return n;
    }
    
    public Attributes toAttributes() {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        for (int i = 0; i < this.entries.size(); ++i) {
            Rdn.RdnEntry rdnEntry = (Rdn.RdnEntry)this.entries.get(i);
            Attribute put = basicAttributes.put(rdnEntry.getType(), rdnEntry.getValue());
            if (put != null) {
                put.add(rdnEntry.getValue());
                basicAttributes.put(put);
            }
        }
        return basicAttributes;
    }
    
    public int size() {
        return this.entries.size();
    }
    
    public static String escapeValue(Object o) {
        return (o instanceof byte[]) ? escapeBinaryValue((byte[])(byte[])o) : escapeStringValue((String)o);
    }
    
    private static String escapeStringValue(String s) {
        char[] charArray = s.toCharArray();
        StringBuilder sb = new StringBuilder(2 * s.length());
        int n;
        for (n = 0; n < charArray.length && isWhitespace(charArray[n]); ++n) {}
        int n2;
        for (n2 = charArray.length - 1; n2 >= 0 && isWhitespace(charArray[n2]); --n2) {}
        for (int i = 0; i < charArray.length; ++i) {
            int n3 = charArray[i];
            if (i < n || i > n2 || ",=+<>#;\"\\".indexOf(n3) >= 0) {
                sb.append('\\');
            }
            sb.append((char)n3);
        }
        return sb.toString();
    }
    
    private static String escapeBinaryValue(byte[] array) {
        StringBuilder sb = new StringBuilder(1 + 2 * array.length);
        sb.append("#");
        for (int i = 0; i < array.length; ++i) {
            int n = array[i];
            sb.append(Character.forDigit(0xF & n >>> 4, 16));
            sb.append(Character.forDigit(0xF & n, 16));
        }
        return sb.toString();
    }
    
    public static Object unescapeValue(String str) {
        char[] charArray;
        int i;
        int length;
        for (charArray = str.toCharArray(), i = 0, length = charArray.length; i < length && isWhitespace(charArray[i]); ++i) {}
        while (i < length && isWhitespace(charArray[length - 1])) {
            --length;
        }
        if (length != charArray.length && i < length && charArray[length - 1] == '\\') {
            ++length;
        }
        if (i >= length) {
            return "";
        }
        if (charArray[i] == '#') {
            return decodeHexPairs(charArray, ++i, length);
        }
        if (charArray[i] == '\"' && charArray[length - 1] == '\"') {
            ++i;
            --length;
        }
        StringBuilder sb = new StringBuilder(length - i);
        int n = -1;
        for (; i < length; ++i) {
            if (charArray[i] == '\\' && i + 1 < length) {
                if (!Character.isLetterOrDigit(charArray[i + 1])) {
                    ++i;
                    sb.append(charArray[i]);
                    n = i;
                }
                else {
                    byte[] utf8Octets = getUtf8Octets(charArray, i, length);
                    if (utf8Octets.length > 0) {
                        try {
                            sb.append(new String(utf8Octets, "UTF8"));
                        }
                        catch (UnsupportedEncodingException ex) {}
                        i += utf8Octets.length * 3 - 1;
                    }
                    else {
                        throw new IllegalArgumentException("Not a valid attribute string value:" + str + ",improper usage of backslash");
                    }
                }
            }
            else {
                sb.append(charArray[i]);
            }
        }
        int length2 = sb.length();
        if (isWhitespace(sb.charAt(length2 - 1)) && n != length - 1) {
            sb.setLength(length2 - 1);
        }
        return sb.toString();
    }
    
    private static byte[] decodeHexPairs(char[] value, int n, int n2) {
        byte[] array = new byte[(n2 - n) / 2];
        int n3 = 0;
        while (n + 1 < n2) {
            int digit = Character.digit(value[n], 16);
            int digit2 = Character.digit(value[n + 1], 16);
            if (digit >= 0) {
                if (digit2 < 0) {
                    break;
                }
                array[n3] = (byte)((digit << 4) + digit2);
                n += 2;
                ++n3;
            }
            else {
                break;
            }
        }
        if (n != n2) {
            throw new IllegalArgumentException("Illegal attribute value: " + new String(value));
        }
        return array;
    }
    
    private static byte[] getUtf8Octets(char[] array, int n, int n2) {
        byte[] array2 = new byte[(n2 - n) / 3];
        int n3 = 0;
        while (n + 2 < n2 && array[n++] == '\\') {
            int digit = Character.digit(array[n++], 16);
            int digit2 = Character.digit(array[n++], 16);
            if (digit >= 0) {
                if (digit2 < 0) {
                    break;
                }
                array2[n3++] = (byte)((digit << 4) + digit2);
            }
            else {
                break;
            }
        }
        if (n3 == array2.length) {
            return array2;
        }
        byte[] array3 = new byte[n3];
        System.arraycopy(array2, 0, array3, 0, n3);
        return array3;
    }
    
    private static boolean isWhitespace(char c) {
        return c == ' ' || c == '\r';
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        objectOutputStream.writeObject(this.toString());
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.entries = new ArrayList(1);
        String str = (String)objectInputStream.readObject();
        try {
            new Rfc2253Parser(str).parseRdn(this);
        }
        catch (InvalidNameException ex) {
            throw new StreamCorruptedException("Invalid name: " + str);
        }
    }
    
    static String access$300(byte[] array) {
        return escapeBinaryValue(array);
    }
    
    private static class RdnEntry implements Comparable<Rdn.RdnEntry>
    {
        private String type;
        private Object value;
        private String comparable;
        
        private RdnEntry() {
            this.comparable = null;
        }
        
        String getType() {
            return this.type;
        }
        
        Object getValue() {
            return this.value;
        }
        
        @Override
        public int compareTo(Rdn.RdnEntry rdnEntry) {
            int compareToIgnoreCase = this.type.compareToIgnoreCase(rdnEntry.type);
            if (compareToIgnoreCase != 0) {
                return compareToIgnoreCase;
            }
            if (this.value.equals(rdnEntry.value)) {
                return 0;
            }
            return this.getValueComparable().compareTo(rdnEntry.getValueComparable());
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof Rdn.RdnEntry)) {
                return false;
            }
            Rdn.RdnEntry rdnEntry = (Rdn.RdnEntry)o;
            return this.type.equalsIgnoreCase(rdnEntry.type) && this.getValueComparable().equals(rdnEntry.getValueComparable());
        }
        
        @Override
        public int hashCode() {
            return this.type.toUpperCase(Locale.ENGLISH).hashCode() + this.getValueComparable().hashCode();
        }
        
        @Override
        public String toString() {
            return this.type + "=" + Rdn.escapeValue(this.value);
        }
        
        private String getValueComparable() {
            if (this.comparable != null) {
                return this.comparable;
            }
            if ((this.value instanceof byte[])) {
                this.comparable = Rdn.access$300((byte[])(byte[])this.value);
            }
            else {
                this.comparable = ((String)this.value).toUpperCase(Locale.ENGLISH);
            }
            return this.comparable;
        }
    }
}
